#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Utilities for virtual broker.

"""

from autotrader.brokers.broker_utils import BrokerUtils

class Utils(BrokerUtils):
    def __init__(self):
        return